import bisect
from collections import defaultdict
from collections import Counter
import copy
import datetime as dt
import fnmatch
from ftplib import FTP
import glob
from io import BytesIO,StringIO
import inspect
import itertools
import linecache
import math
import matplotlib
import matplotlib.pyplot as plt
import multiprocessing as mp
from natsort import natsorted, ns
import numpy as np
import numpy as npaa
import operator
import os 
from os import remove, close
import pandas as df
import pandas as pd
import pickle
import scipy
### desactivated MAYOBS25 230925
#from scipy.signal import butter, lfilter, freqz
import shutil
import string
import struct
import sys
import subprocess
import re
import tempfile
from tempfile import mkstemp
import tabulate
import time
import urllib
import uuid
