#!/usr/bin/env python3
# -*- coding: utf-8 -*-

##### megalib must be imported with
##### from geodezyx.megalib.megalib import *

import geodezyx.megalib.geoclass as gcls
import geodezyx.megalib.geodetik as geok
import geodezyx.megalib.genefun  as genefun
import geodezyx.megalib.genefun  as gf
import geodezyx.megalib.softs_runner  as softs_runner
import geodezyx.megalib.geo_files_converter_lib  as gfc



